#include "rfits.h"

SEXP rfits_get_num_hdus(SEXP filename)
{
  short prot=0;
  int hdunum,err,status = 0;
  char* fname;
  fitsfile* fp;
  SEXP numhdus;

  fname = CHAR(STRING_ELT(filename,0));
  err = fits_open_file(&fp,fname,READONLY,&status);
  if ( status != 0 ) return R_NilValue;

  err = fits_get_num_hdus(fp,&hdunum,&status);

  PROTECT(numhdus = allocVector(INTSXP,1)); prot++;
  INTEGER(numhdus)[0] = hdunum;

  err = fits_close_file(fp,&status);
  UNPROTECT(prot);
  return(numhdus);
}

SEXP rfits_get_hdu_type(SEXP filename,SEXP hdu)
{
  short prot=0;
  int hdutype,err,status = 0;
  int* hduint;
  char* fname;
  char hduchar[7];
  fitsfile* fp;
  SEXP retval;

  fname = CHAR(STRING_ELT(filename,0));
  err = fits_open_file(&fp,fname,READONLY,&status);
  if ( status != 0 ) return R_NilValue;
  PROTECT(hdu = coerceVector(hdu,INTSXP)); prot++;
  hduint = INTEGER(hdu);
  err = fits_movabs_hdu(fp,hduint[0],&hdutype,&status);
  if ( status != 0 ) { UNPROTECT(prot); return R_NilValue; }
  if ( hdutype == IMAGE_HDU ) {
    strcpy(hduchar,"image");
  } else if ( hdutype == ASCII_TBL ) {
    strcpy(hduchar,"ascii");
  } else if ( hdutype == BINARY_TBL ) {
    strcpy(hduchar,"binary");
  } else if ( hdutype == ANY_HDU ) {
    strcpy(hduchar,"any");
  }
  PROTECT(retval = allocVector(STRSXP,1)); prot++;
  SET_STRING_ELT(retval,0,mkChar(hduchar));
  err = fits_close_file(fp,&status);
  UNPROTECT(prot);
  return(retval);
}

SEXP rfits_read_image(SEXP filename,SEXP hdu)
{
  short prot=0;
  int ii;
  int err,status = 0;
  int bitpix,naxis,anynull;
  int hdutype;
  int* hduint;
  long imgsize = 1;
  long* naxes;
  long* fpixl;
  double* imgdatadbl;
  char* fname;
  fitsfile* fp;
  SEXP imgdata,dim;

  fname = CHAR(STRING_ELT(filename,0));
  err = fits_open_file(&fp,fname,READONLY,&status);
  if ( status != 0 ) return R_NilValue;
  PROTECT(hdu = coerceVector(hdu,INTSXP)); prot++;
  hduint = INTEGER(hdu);
  err = fits_movabs_hdu(fp,hduint[0],&hdutype,&status);
  if ( hdutype != IMAGE_HDU || status != 0 ) {
    printf("RFITS ERROR: the hdu does not contain a FITS image.\n");
    UNPROTECT(prot); return R_NilValue;
  }
  err = fits_get_img_type(fp,&bitpix,&status);
  if ( status != 0 ) { UNPROTECT(prot); return R_NilValue; }
  err = fits_get_img_dim(fp,&naxis,&status);
  if ( status != 0 || naxis <= 0 ) { UNPROTECT(prot); return R_NilValue; }
  naxes = (long*)malloc(sizeof(long)*naxis);
  fpixl = (long*)malloc(sizeof(long)*naxis);
  err = fits_get_img_size(fp,naxis,naxes,&status);
  for ( ii = 0 ; ii < naxis ; ii++ ) {
    imgsize *= naxes[ii];
    fpixl[ii] = 1;
  }

  PROTECT(imgdata = allocVector(REALSXP,imgsize)); prot++;
  imgdatadbl = REAL(imgdata);

  PROTECT(dim = allocVector(INTSXP,naxis)); prot++;
  for ( ii = 0 ; ii < naxis ; ii++ ) {
    INTEGER(dim)[ii] = (int)(naxes[ii]);
  }

  if ( bitpix == 8 ) {
    {
      short nullval = -999;
      short* data;
      data = (short*)malloc(sizeof(short)*imgsize);
      err = fits_read_pix(fp,TSHORT,fpixl,imgsize,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { free(data); UNPROTECT(prot); return R_NilValue; }
      for ( ii = 0 ; ii < imgsize ; ii++ ) {
        imgdatadbl[ii] = (double)(data[ii]);
      }
      free(data);
    }
  } else if ( bitpix == 16 ) {
    {
      int nullval = -999;
      int* data;
      data = (int*)malloc(sizeof(int)*imgsize);
      err = fits_read_pix(fp,TINT,fpixl,imgsize,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { free(data); UNPROTECT(prot); return R_NilValue; }
      for ( ii = 0 ; ii < imgsize ; ii++ ) {
        imgdatadbl[ii] = (double)(data[ii]);
      }
      free(data);
    }
  } else if ( bitpix == 32 ) {
    {
      long nullval = -999;
      long* data;
      data = (long*)malloc(sizeof(long)*imgsize);
      err = fits_read_pix(fp,TLONG,fpixl,imgsize,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { free(data); UNPROTECT(prot); return R_NilValue; }
      for ( ii = 0 ; ii < imgsize ; ii++ ) {
        imgdatadbl[ii] = (double)(data[ii]);
      }
      free(data);
    }
  } else if ( bitpix == -32 ) {
    {
      float nullval = -999.0;
      float* data;
      data = (float*)malloc(sizeof(float)*imgsize);
      err = fits_read_pix(fp,TFLOAT,fpixl,imgsize,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { free(data); UNPROTECT(prot); return R_NilValue; }
      for ( ii = 0 ; ii < imgsize ; ii++ ) {
        imgdatadbl[ii] = (double)(data[ii]);
      }
      free(data);
    }
  } else if ( bitpix == -64 ) {
    {
      double nullval = -999.0;
      double* data;
      data = (double*)malloc(sizeof(double)*imgsize);
      err = fits_read_pix(fp,TDOUBLE,fpixl,imgsize,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { free(data); UNPROTECT(prot); return R_NilValue; }
      for ( ii = 0 ; ii < imgsize ; ii++ ) {
        imgdatadbl[ii] = data[ii];
      }
      free(data);
    }
  } else {
    printf("RFITS ERROR: unrecognized data type (bitpix = %d)\n",bitpix);
    UNPROTECT(prot); return R_NilValue;
  }
  setAttrib(imgdata,R_DimSymbol,dim);
  err = fits_close_file(fp,&status);
  UNPROTECT(prot);
  return(imgdata);
}

SEXP rfits_get_col_names(SEXP filename,SEXP hdu)
{
  short prot=0;
  int ii,colnum,ncols,hdutype,err,status = 0;
  int* hduint;
  char* fname;
  char colname[9];
  fitsfile* fp;
  SEXP colnames;

  fname = CHAR(STRING_ELT(filename,0));
  err = fits_open_file(&fp,fname,READONLY,&status);
  if ( status != 0 ) return R_NilValue;
  PROTECT(hdu = coerceVector(hdu,INTSXP)); prot++;
  hduint = INTEGER(hdu);
  err = fits_movabs_hdu(fp,hduint[0],&hdutype,&status);
  if ( status != 0 ) { UNPROTECT(prot); return R_NilValue; }
  if ( hdutype != BINARY_TBL && hdutype != ASCII_TBL ) {
    UNPROTECT(prot); return R_NilValue; 
  }

  err = fits_get_num_cols(fp,&ncols,&status);
  PROTECT(colnames = allocVector(STRSXP,ncols)); prot++;

  status = 0; ii = 0;
  while ( status != COL_NOT_FOUND ) {
    err = fits_get_colname(fp,CASEINSEN,"*",colname,&colnum,&status);
    SET_STRING_ELT(colnames,ii,mkChar(colname));
    ii++;
  }
  err = fits_close_file(fp,&status);
  UNPROTECT(prot);
  return(colnames);
}

SEXP rfits_read_col(SEXP filename,SEXP column,SEXP cwidth,SEXP hdu,
		SEXP ischar)
{
  short prot=0;
  int err,status=0;
  int hdutype,anynull,typecode;
  long ii,nrows,width,repeat;
  int* colint=NULL;
  int colwidth=1;
  int* hduint;
  int* ischarint;
  double* coldatadbl;
  char* fname;
  char* colchar=NULL;
  fitsfile* fp=NULL;
  SEXP coldata;

  fname = CHAR(STRING_ELT(filename,0));
  err = fits_open_file(&fp,fname,READONLY,&status);
  if ( status != 0 ) { read_col_clean(prot,fp,colint); return R_NilValue; }

  PROTECT(hdu = coerceVector(hdu,INTSXP)); prot++;
  hduint = INTEGER(hdu);
  err = fits_movabs_hdu(fp,hduint[0],&hdutype,&status);
  if ( hdutype != BINARY_TBL || status != 0 ) {
    printf("RFITS ERROR: the hdu does not contain a FITS binary table.\n");
    read_col_clean(prot,fp,colint); return R_NilValue;
  }

  PROTECT(cwidth = coerceVector(cwidth,INTSXP)); prot++;
  colwidth = INTEGER(cwidth)[0];

  PROTECT(ischar = coerceVector(ischar,INTSXP)); prot++;
  ischarint = INTEGER(ischar);
  if ( ischarint[0] == 0 ) {
    PROTECT(column = coerceVector(column,INTSXP)); prot++;
    colint = (int*)malloc(sizeof(int));
    colint[0] = INTEGER(column)[0];
  } else {
    colchar = CHAR(STRING_ELT(column,0));
    colint = (int*)malloc(sizeof(int));
    err = fits_get_colnum(fp,CASEINSEN,colchar,colint,&status);
    if ( status != 0 ) {
      printf("RFITS ERROR: the named column not found.\n");
      read_col_clean(prot,fp,colint); return R_NilValue;
    }
  }

  err = fits_get_num_rows(fp,&nrows,&status);
  if ( status != 0 || nrows == 0 ) { 
    read_col_clean(prot,fp,colint); return R_NilValue;
  }

  err = fits_get_coltype(fp,colint[0],&typecode,&repeat,&width,&status);

  if ( typecode != TSTRING ) {
    PROTECT(coldata = allocVector(REALSXP,nrows*colwidth)); prot++;
    coldatadbl = REAL(coldata);
  } else {
    PROTECT(coldata = allocVector(STRSXP,nrows*colwidth)); prot++;
  }

  if ( typecode == TSHORT ) {
    {
      short nullval = -99;
      short* data;
      data = (short*)malloc(sizeof(short)*nrows*colwidth);
      err = fits_read_col(fp,TSHORT,colint[0],1,1,nrows*colwidth,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { 
        free(data); read_col_clean(prot,fp,colint); return R_NilValue;
      }
      for ( ii = 0 ; ii < nrows*colwidth ; ii++ ) {
        coldatadbl[ii] = (double)(data[ii]);
      }
      free(data);
    }
  } else if ( typecode == TINT ) {
    {
      int nullval = -99;
      int* data;
      data = (int*)malloc(sizeof(int)*nrows*colwidth);
      err = fits_read_col(fp,TINT,colint[0],1,1,nrows*colwidth,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { 
        free(data); read_col_clean(prot,fp,colint); return R_NilValue;
      }
      for ( ii = 0 ; ii < nrows*colwidth ; ii++ ) {
        coldatadbl[ii] = (double)(data[ii]);
      }
      free(data);
    }
  } else if ( typecode == TLONG ) {
    {
      long nullval = -99;
      long* data;
      data = (long*)malloc(sizeof(long)*nrows*colwidth);
      err = fits_read_col(fp,TLONG,colint[0],1,1,nrows*colwidth,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { 
        free(data); read_col_clean(prot,fp,colint); return R_NilValue;
      }
      for ( ii = 0 ; ii < nrows*colwidth ; ii++ ) {
        coldatadbl[ii] = (double)(data[ii]);
      }
      free(data);
    }
  } else if ( typecode == TFLOAT ) {
    {
      float nullval = -99.0;
      float* data;
      data = (float*)malloc(sizeof(float)*nrows*colwidth);
      err = fits_read_col(fp,TFLOAT,colint[0],1,1,nrows*colwidth,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { 
        free(data); read_col_clean(prot,fp,colint); return R_NilValue;
      }
      for ( ii = 0 ; ii < nrows*colwidth ; ii++ ) {
        coldatadbl[ii] = (double)(data[ii]);
      }
      free(data);
    }
  } else if ( typecode == TDOUBLE ) {
    {
      double nullval = -99.0;
      double* data;
      data = (double*)malloc(sizeof(double)*nrows*colwidth);
      err = fits_read_col(fp,TDOUBLE,colint[0],1,1,nrows*colwidth,&nullval,data,
                          &anynull,&status);
      if ( status != 0 ) { 
        free(data); read_col_clean(prot,fp,colint); return R_NilValue;
      }
      for ( ii = 0 ; ii < nrows*colwidth ; ii++ ) {
        coldatadbl[ii] = data[ii];
      }
      free(data);
    }
  } else if ( typecode == TSTRING ) {
    {
      char** data;
      data = (char**)malloc(sizeof(char*)*nrows*colwidth);
      for ( ii = 0 ; ii < nrows*colwidth ; ii++ ) {
        data[ii] = (char*)malloc(80*sizeof(char));
      }
      err = fits_read_col(fp,TSTRING,colint[0],1,1,nrows*colwidth,0,data,
                          &anynull,&status);
      if ( status != 0 ) { 
        for ( ii = 0 ; ii < nrows*colwidth ; ii++ ) {
          free(data[ii]);
        }
        free(data); read_col_clean(prot,fp,colint); return R_NilValue;
      }
      for ( ii = 0 ; ii < nrows*colwidth ; ii++ ) {
	SET_STRING_ELT(coldata,ii,mkChar(data[ii]));
      }
      for ( ii = 0 ; ii < nrows*colwidth ; ii++ ) {
        free(data[ii]);
      }
      free(data);
    }
  } else {
    printf("RFITS ERROR: unrecognized data type (typecode = %d).\n",typecode);
    read_col_clean(prot,fp,colint); return R_NilValue;
  }

  read_col_clean(prot,fp,colint);

  return(coldata);
}

void read_col_clean(short prot,fitsfile* fp,int* colint)
{
  int err,status=0;
  free(colint);
  if ( fp != NULL ) err = fits_close_file(fp,&status);
  if ( prot > 0 ) UNPROTECT(prot);
}

SEXP rfits_read_key(SEXP filename,SEXP keyword,SEXP hdu)
{
  short prot=0;
  int err,status=0;
  int hdutype;
  int* hduint;
  char* fname;
  char* kname;
  char kword[80];
  char comment[80];
  fitsfile* fp=NULL;
  SEXP keyvalue;

  fname = CHAR(STRING_ELT(filename,0));
  err = fits_open_file(&fp,fname,READONLY,&status);
  if ( status != 0 ) { read_key_clean(prot,fp); return R_NilValue; }

  PROTECT(hdu = coerceVector(hdu,INTSXP)); prot++;
  hduint = INTEGER(hdu);
  err = fits_movabs_hdu(fp,hduint[0],&hdutype,&status);

  kname = CHAR(STRING_ELT(keyword,0));
  err = fits_read_keyword(fp,kname,kword,comment,&status);
  if ( status != 0 ) { read_key_clean(prot,fp); return R_NilValue; }
  PROTECT(keyvalue = allocVector(STRSXP,1)); prot++;
  SET_STRING_ELT(keyvalue,0,mkChar(kword));
  read_key_clean(prot,fp);
  return(keyvalue);
}

void read_key_clean(short prot,fitsfile* fp)
{
  int err,status=0;
  if ( fp != NULL ) err = fits_close_file(fp,&status);
  if ( prot > 0 ) UNPROTECT(prot);
}

